# -*- ruby -*-
# vim: set nosta noet ts=4 sw=4 ft=ruby:

require 'inversion'
require 'loggability'

class Roda
	module RodaPlugins

		# A templated content-generation plugin for Roda Apps. It uses the
		# Inversion[http://deveiate.org/projects/Inversion] templating system.
		#
		# It adds:
		#
		# * a preloaded/cached template table
		# * a mechanism for fetching templates from the table
		# * a global layout template which is automatically wrapped around responses
		#
		# Note: This is based almost entirely off the Strelka plugin of identical
		# function.
		#
		# == Usage
		#
		# To use it, just load the <tt>:inversion</tt> plugin in your app:
		#
		#   plugin :inversion, '/path/to/templates/directory'
		#
		# You can specify more than one templates directory; they'll be searched in the order
		# you specify for the named template and the first match wins:
		#
		#   plugin :inversion '/more/specific/templates', '/generic/app/templates'
		#
		# Then declare one or more templates that your application will use:
		#
		#   templates console:   'views/console.tmpl',
		#             proctable: 'partials/proctable.tmpl'
		#
		# Then, inside your app, you can fetch a copy of one or more of the templates and
		# return it as the response:
		#
		#   r.root do
		#       tmpl = template( :console )
		#       tmpl.message = "Everything's up."
		#       tmpl.proctable = proctable
		#       return tmpl
		#   end
		#
		#
		# You can also just return the template if you don't need to do anything else to the
		# response.
		#
		# When returning a template, either in the body of the response or directly, it will
		# automatically set a few attributes for commonly-used objects:
		#
		# request ::          The current Roda::Request
		# response ::         The current Roda::Response
		# app ::              The application object (Roda::App instance).
		#
		# If your app will *only* be loading and returning a template without doing anything
		# with it, you can return just its name:
		#
		#   r.root do
		#       return :console
		#   end
		#
		# It will be loaded, set as the response body, and the above common objects added to it.
		#
		#
		# === Layouts
		#
		# Very often, you'll want all or most of the views in your app to share a common page
		# layout. To accomplish this, you can declare a layout template:
		#
		#   layout 'layout.tmpl'
		#
		# Any template that you return will be set as the 'body' attribute of this layout
		# template (which you'd place into the layout with <tt><?attr body ?></tt>) and the
		# layout rendered as the body of the response.
		#
		# Note that if you want any of the "common objects" from above with a layout template,
		# they'll be set on it since it's the top-level template, but you can still access them
		# using the <tt><?import ?></tt> directive:
		#
		#   <?import request, app ?>
		#
		#
		# == Template Locations
		#
		# Inversion looks for templates in a load path much like Ruby does for libraries that
		# you 'require'. It contains just the current working directory by default. You can add
		# your own template directories via the config file (under +template_paths+ in
		# the +templates+ section), or programmatically from your application.
		#
		module Inversion

			### Add the specified +template_paths+ to Inversion.
			def self::configure( app, *template_paths )
				::Inversion::Template.template_paths.concat( template_paths )
			end


			module ClassMethods

				### Add this Gem's template location to the Inversion search path.
				def self::extended( subclass )
					subclass.instance_variable_set( :@template_map, {} )
					subclass.instance_variable_set( :@layout_template, nil )

					super
				end


				# The map of template names to template file paths.
				attr_reader :template_map

				# An optional layout/wrap template.
				attr_accessor :layout_template


				### Load instances for all the template paths specified in the App's class
				### and return them in a hash keyed by name (Symbol).
				def templates( hash )
					@template_map = hash.each_with_object( {} ) do |(name, path), map|
						enc = Encoding.default_internal || Encoding::UTF_8
						map[ name ] = ::Inversion::Template.load( path, encoding: enc )
						map
					end

					return self.template_map
				end


				### Load an Inversion::Template for the layout template.
				def layout( tmpl )
					enc = Encoding.default_internal || Encoding::UTF_8
					self.layout_template = ::Inversion::Template.load( tmpl, encoding: enc )
					return self.layout_template
				end
			end


			module InstanceMethods

				### Return a previously declared template object.
				def template( tmpl )
					template = self.class.template_map[ tmpl ] or
						raise ArgumentError, "no %p template registered!" % [ tmpl ]
					template.reload if template.changed?
					return template.dup
				end
			end


			module RequestMethods

				### Hook block results for template rendering.
				def block_result_body( result )
					if ( template = self.extract_template_from_result( result ) )

						# Wrap the template in a layout if there is one
						template = self.wrap_in_layout( template )

						# Set some default stuff on the top-level template
						self.set_common_attributes( template )

						return template.render
					else
						super
					end
				end


				#########
				protected
				#########

				### Fetch the template from the +result+ (if there is one) and return it.
				def extract_template_from_result( result )
					templates = self.scope.class.template_map

					# Result is a template name.
					if result.is_a?( Symbol ) && templates.key?( result )
						return scope.template( result )

						# Result is a template object.
					elsif result.respond_to?( :render )
						return result
					end

					# Not templated; returned as-is
					return nil
				end


				### Wrap the specified +content+ template in the layout template and
				### return it. If there isn't a layout declared, just return +content+ as-is.
				def wrap_in_layout( content )
					layout = self.scope.class.layout_template or return content

					layout.reload if layout.changed?
					l_template = layout.dup
					l_template.body = content

					return l_template
				end


				### Set some default values from the request in the given top-level +template+.
				def set_common_attributes( template )
					template.request  = self
					template.response = self.scope.response
					template.app      = self.scope
				end

			end

		end # module Inversion


		# Hook: Inform Roda that we exist.
		register_plugin :inversion, Inversion

	end # module RodaPlugins
end # class Roda

